from typing import Optional
from pyexlatex.models.format.text.color.deftypes.base import ColorDefinition


class RGB(ColorDefinition):
    definition_type = 'RGB'

    def __init__(self, red: int = 255, green: int = 255, blue: int = 255, color_name: Optional[str] = None):
        self.red = red
        self.green = green
        self.blue = blue
        color_content = ','.join([str(color) for color in [red, green, blue]])
        super().__init__(color_content=color_content, color_name=color_name)



from pyexlatex.models.control.documentclass.classtypes.documentclasstype import DocumentClassType

BUILTIN_CLASS_TYPE_NAMES = [
    'article',
    'report',
    'letter',
    'book',
    'proc',
    'slides',
    'beamer',
    'standalone',
]

BUILTIN_CLASS_TYPES = {style_str: DocumentClassType(style_str) for style_str in BUILTIN_CLASS_TYPE_NAMES}



from typing import Sequence, Optional
from pyexlatex.models.item import Item
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.graphics.tikz.item import TikzOptionHandler


class Scope(TextAreaMixin, Item, TikzOptionHandler):
    name = 'scope'

    def __init__(self, contents, options: Optional[Sequence[str]] = None):
        self.options = options
        super().__init__(self.name, contents, env_modifiers=self.options_str)




from typing import Optional, Dict
from pyexlatex.models.references.bibtex.base import BibTexEntryBase


class BibTexMisc(BibTexEntryBase):
    """
    Biblography document which is of miscellaneous type.
    """
    item_type = 'misc'
    optional_attrs = ['author', 'title', 'howpublished', 'month', 'year', 'note']

    def __init__(self, item_accessor: str, author: Optional[str] = None, title: Optional[str] = None,
                 howpublished: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None,
                 note: Optional[str] = None):
        self.author = author
        self.title = title
        self.howpublished = howpublished
        self.year = year
        self.month = month
        self.note = note

        self._validate()

        super().__init__(item_accessor)

    def _validate(self):
        all_values = [getattr(self, attr) for attr in self.optional_attrs]
        if all([value is None for value in all_values]):
            raise ValueError('must pass something non-None to BibTexMisc, got all None')



from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.presentation.beamer.overlay.range import Range
from pyexlatex.presentation.beamer.overlay.next import NextWithIncrement, NextWithoutIncrement
from pyexlatex.presentation.beamer.overlay.until_end import UntilEnd
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.presentation.beamer.block import Block, AlertBlock, ExamplesBlock
from pyexlatex.presentation.presentation import Presentation
from pyexlatex.presentation.beamer.templates.frames.grid import GridFrame
from pyexlatex.presentation.beamer.templates.lists.dim_reveal_items import DimAndRevealListItems, DimAndRevealListItem
from pyexlatex.presentation.beamer.templates.frames.two_col import (
    TwoColumnGraphicDimRevealFrame,
    TwoColumnGraphicFrame,
    BasicTwoColumnGraphicFrame,
    BasicTwoColumnFrame,
    TwoColumnFrame
)
from pyexlatex.presentation.beamer.templates.frames.dim_reveal import DimRevealListFrame
from pyexlatex.presentation.beamer.templates.frames.graphic import (
    GraphicFrame,
    MultiGraphicFrame,
)
from pyexlatex.logic.format.sizing import adjust_to_full_size_and_center
from pyexlatex.presentation.beamer.templates.appendix import Appendix



from pyexlatex.models.item import Item
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.logic.format.contents import format_contents


class Abstract(DocumentItem, Item):
    name = 'abstract'

    def __init__(self, content):
        content = format_contents(content)
        super().__init__(self.name, content)



from typing import Sequence, Tuple, Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay import Overlay
    from pyexlatex.graphics.tikz import Node
from pyexlatex.graphics.tikz.path import SpecificPath


class Arrow(SpecificPath):
    """
    Draws an arrow from one location to another.
    """
    draw_type = '--'
    path_type = 'draw'

    def __init__(self, from_: Union[Tuple[int, int], 'Node'], to: Union[Tuple[int, int], 'Node'],
                 double_sided: bool = False,
                 options: Optional[Sequence[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.from_ = from_
        self.to = to
        self.double_sided = double_sided

        options = self._get_list_copy_from_list_or_none(options)
        options.append(self.arrow_option)

        super().__init__((from_, to), options=options, overlay=overlay)

    @property
    def arrow_option(self) -> str:
        if self.double_sided:
            return '<->'
        return '->'



import re

from mixins.repr import ReprMixin
from pyexlatex.models.item import ItemBase
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.texgen.packages.columntypes import ColumnTypesPackage


class ColumnAlignment(ReprMixin, ItemBase):
    repr_cols = ['align']

    def __init__(self, align_str: str):
        self._validate_align_str(align_str)
        self.align = align_str

    def __str__(self):
        return self.align

    def __add__(self, other):
        return self.align + other.align

    def __radd__(self, other):
        return other.align + self.align

    def _validate_align_str(self, align_str):
        basic_pattern = re.compile(r'[lcr|]')
        length_pattern = re.compile(r'[LCRD]\{[\d\w\s.]+\}')

        basic_match = basic_pattern.fullmatch(align_str)
        length_match = length_pattern.fullmatch(align_str)

        if length_match:
            self._add_requirements_for_length_match()

        if not (basic_match or length_match):
            raise ValueError(f'expected alignment of l, c, r, |, L{{size}}, C{{size}}, R{{size}}, or D{{size}}. Got {align_str}')

    def _add_requirements_for_length_match(self):
        self.add_package(ColumnTypesPackage())


class ColumnsAlignment(ReprMixin, ContainerItem):
    repr_cols = ['aligns']

    def __init__(self, aligns: [ColumnAlignment]= None, num_columns: int=None):
        self.aligns = ColumnsAlignment._get_aligns(aligns, num_columns)
        self.add_data_from_content(self.aligns)

    def __str__(self):
        return ''.join(str(align) for align in self.aligns)

    def __iter__(self):
        for align in self.aligns:
            yield align

    @staticmethod
    def _get_aligns(aligns: [ColumnAlignment]= None, num_columns: int=None):
        if aligns is None and num_columns is None:
            raise ValueError('must pass aligns or num columns')

        # default align is first column left, rest centered
        if aligns is None and num_columns is not None:
            return [ColumnAlignment('l')] + [ColumnAlignment('c')] * (num_columns - 1)

        # if we don't know how many columns, must assume passed number of aligns is correct
        if num_columns is None:
            return aligns

        # number of alignments matches number of columns. no extra processing needed
        if len(aligns) == num_columns:
            return aligns

        # if one alignment is passed with many columns, use that align with all columns
        if len(aligns) == 1:
            return [aligns[0]] * num_columns
        else:
            raise ValueError(f'got {len(aligns)} alignments for {num_columns} columns. unclear how to apply')

    @classmethod
    def from_alignment_str(cls, align_str: str, num_columns: int=None):
        align_str_list = _full_align_str_to_align_str_list(align_str)
        aligns = [ColumnAlignment(align) for align in align_str_list]

        return cls(aligns, num_columns=num_columns)


def _full_align_str_to_align_str_list(align_str: str):
    split_letters = ['l', 'c', 'r', '|', 'L', 'C', 'R']
    out_list = []
    collected_letters = ''
    split = True

    for letter in align_str:
        # beginning inside of length str. don't split while inside
        if letter == '{':
            split = False
        # end of inside of length str. turn splitting back on
        if letter == '}':
            split = True
        # if splitting, output what we've got so far and start a new item
        if split and letter in split_letters:
            out_list.append(collected_letters)
            collected_letters = ''
        # if not splitting, add to current item
        collected_letters += letter

    # Clean up list from loop. Remove first (always blank), and add last
    out_list.append(collected_letters)
    out_list = out_list[1:]

    return out_list


from pyexlatex.models.item import MultiOptionSimpleItem


class ReNewCommand(MultiOptionSimpleItem):
    name = 'renewcommand'

    def __init__(self, command_name: str, command_definition: str):
        super().__init__(self.name, rf'\{command_name}', command_definition)





from pyexlatex.table.models.spacing.cell import CellSpacer
from pyexlatex.table.models.data.row import DataRow
from pyexlatex.table.models.data.valuestable import ValuesTable

class RowPadTable(ValuesTable):

    def __init__(self, num_cols: int = 1):
        row = DataRow([CellSpacer(num_cols)])
        super().__init__([row])

    @property
    def num_columns(self):
        return len(self.rows)

    def __repr__(self):
        return f'<RowPadTable({self.num_columns})>'






from pyexlatex.models.item import SimpleItem


class Subtitle(SimpleItem):
    name = 'subtitle'

    def __init__(self, subtitle: str):
        self.subtitle = subtitle
        super().__init__(self.name, self.subtitle)


from pyexlatex.models.item import NoOptionsNoContentsItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.format.breaks import LineBreak


class NoIndent(TextAreaMixin, NoOptionsNoContentsItem):
    """
    Do not have an indent for the current paragraph
    """
    name = 'noindent'

    def __init__(self, contents, **kwargs):
        TextAreaMixin.__init__(self, self.name, contents)
        NoOptionsNoContentsItem.__init__(self, self.name, **kwargs)
        if not isinstance(self.contents, (list, tuple)):
            self.contents = [self.contents]

    def __str__(self):
        from pyexlatex.logic.builder import _build
        from pyexlatex.texgen import no_options_no_contents_str
        if isinstance(self.contents, str):
            return self.contents
        # Ending line break is necessary to end paragraph, so that next item doesn't have no indent
        return no_options_no_contents_str(self.name, overlay=self.overlay) + _build(self.contents + [LineBreak()])



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.graphics.tikz import Node
from pyexlatex.models.item import ItemBase
from pyexlatex.models.package import Package
from pyexlatex.graphics.tikz.library import TikZLibrary


class DirectionBase(ItemBase):
    """
    Base class for relative positioning directions in TikZ, e.g. above, above right, etc.
    """

    name = '<invalid, do not use DirectionBase directly>'
    is_vertical_direction = None

    def __init__(self, direction: Optional['DirectionBase'] = None, of: Optional['Node'] = None,
                 by: Optional[float] = None, combined_direction: bool = True):
        """

        :param direction:
        :param of:
        :param by:
        :param combined_direction: In some usages of directions, they are specified together, e.g.
            above right=2cm and 3cm. In other usages, they are specified separately, e.g. above=2cm, right=3cm.
            When combined_direction=True, will use the former style, and when False will use the latter style
        """
        self.child_direction = direction
        self.of = of
        self.by = by
        self.combined_direction = combined_direction
        self._validate()
        self._set_vertical_and_horizontal_by()

        self.init_data()
        self.data.packages.extend([
            Package('tikz'),
            TikZLibrary('positioning')
        ])

    def _validate(self):
        if self.child_direction is not None and self.of is not None:
            raise ValueError('must only pass of when there are no other child directions')
        if self.is_vertical_direction is None:
            raise ValueError('must set is_vertical_direction to True or False when subclassing DirectionBase')
        if self.child_direction is not None:
            if self.is_vertical_direction and self.child_direction.is_vertical_direction:
                raise ValueError('cannot combine vertical directions, must be a vertical and horizontal direction')
            if not self.is_vertical_direction and not self.child_direction.is_vertical_direction:
                raise ValueError('cannot combine horizontal directions, must be a vertical and horizontal direction')
            if self.of is not None and self.child_direction.of is not None:
                raise ValueError('cannot pass two of in nested directions')

    @property
    def full_direction(self) -> str:
        if self.child_direction is None:
            return f'{self.name}{self.of_str}'

        if self.is_vertical_direction:
            vertical_name = self.name
            horizontal_name = self.child_direction.name
        else:
            vertical_name = self.child_direction.name
            horizontal_name = self.name

        if self.has_combined_direction:
            return f'{vertical_name} {horizontal_name}{self.of_str}'

        # Handle non-combined direction

        return f'{vertical_name}{self.vertical_of_str}, {horizontal_name}{self.horizontal_of_str}'

    @property
    def has_combined_direction(self) -> bool:
        if self.child_direction is None:
            return self.combined_direction

        if not self.combined_direction or not self.child_direction.combined_direction:
            # Having False in either parent or child means entire is treated as False, since True is the default
            return False

        return True

    @property
    def of_str(self) -> str:
        if self.of is None and (self.child_direction is None or self.child_direction.of is None):
            if not self.by_str:
                return ''
            else:
                return f'={self.by_str}'

        return f'={self.by_str}of {self.of_label}'

    @property
    def vertical_of_str(self) -> Optional[str]:
        if self.has_combined_direction:
            return None

        if self.of is None and (self.child_direction is None or self.child_direction.of is None):
            if not self.vertical_by_str:
                return ''
            else:
                return f'={self.vertical_by_str}'

        return f'={self.vertical_by_str}of {self.of_label}'

    @property
    def horizontal_of_str(self) -> Optional[str]:
        if self.has_combined_direction:
            return None

        if self.of is None and (self.child_direction is None or self.child_direction.of is None):
            if not self.horizontal_by_str:
                return ''
            else:
                return f'={self.horizontal_by_str}'

        return f'={self.horizontal_by_str}of {self.of_label}'

    @property
    def of_label(self) -> Optional[str]:
        if self.of is None and (self.child_direction is None or self.child_direction.of is None):
            return None

        if self.of is not None:
            label = self.of.label
        else:
            label = self.child_direction.of.label
        return label


    @property
    def by_str(self) -> str:
        if self.by is None and (self.child_direction is None or self.child_direction.by is None):
            return ''

        out_str = ''
        if self.vertical_by is not None:
            out_str += self.vertical_by_str
        if self.horizontal_by is not None:
            out_str += self.horizontal_by_str

        return out_str

    @property
    def vertical_by_str(self) -> Optional[str]:
        if self.vertical_by is None:
            return None
        return f'{self.vertical_by}cm '

    @property
    def horizontal_by_str(self) -> Optional[str]:
        if self.horizontal_by is None:
            return None
        horizontal_str = f'{self.horizontal_by}cm '
        if self.has_combined_direction:
            horizontal_str = 'and ' + horizontal_str
        return horizontal_str

    def _set_vertical_and_horizontal_by(self):
        if self.by is None and (self.child_direction is None or self.child_direction.by is None):
            self.vertical_by = None
            self.horizontal_by = None
            return

        vertical_by, horizontal_by = None, None
        if _is_vertical_direction_and_has_by(self):
            vertical_by = self.by
        elif self.child_direction is not None and _is_vertical_direction_and_has_by(self.child_direction):
            vertical_by = self.child_direction.by

        if _is_horizontal_direction_and_has_by(self):
            horizontal_by = self.by
        elif self.child_direction is not None and _is_horizontal_direction_and_has_by(self.child_direction):
            horizontal_by = self.child_direction.by

        self.vertical_by = vertical_by
        self.horizontal_by = horizontal_by

    def __str__(self) -> str:
        return self.full_direction


class VerticalDirection(DirectionBase):
    is_vertical_direction = True


class HorizontalDirection(DirectionBase):
    is_vertical_direction = False


class Above(VerticalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed above the first node.
    """
    name = 'above'


class Right(HorizontalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed to the right of the first node.
    """
    name = 'right'


class Below(VerticalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed below the first node.
    """
    name = 'below'


class Left(HorizontalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed to the left of the first node.
    """
    name = 'left'


def _is_vertical_direction_and_has_by(direction: 'DirectionBase') -> bool:
    return direction.is_vertical_direction and direction.by is not None


def _is_horizontal_direction_and_has_by(direction: 'DirectionBase') -> bool:
    return not direction.is_vertical_direction and direction.by is not None


from pyexlatex.presentation.beamer.overlay.overlay_param import OverlayParameter


class UntilEnd(OverlayParameter):
    """
    Option to be passed to Overlay that makes object exist until the end of the frame
    """

    def __init__(self, content):
        self.content = content

    def __str__(self):
        return f'{self.content}-'



from pyexlatex.models.item import NoOptionsNoContentsItem

LATEX_SIZES = [
    'tiny',
    'scriptsize',
    'footnotesize',
    'small',
    'normalsize',
    'large',
    'Large',
    'LARGE',
    'huge',
    'Huge'
]

class TextSize(NoOptionsNoContentsItem):

    def __init__(self, relative_size_int: int):
        """

        Args:
            relative_size_int: 0 gets the normal font size, negative numbers get smaller sizes, positive numbers get
            larger sizes. Range is from -4 to 5
        """
        self.relative_size_int = relative_size_int
        self._validate()
        self.name = latex_size_str_for_relative_size_int(self.relative_size_int)
        super().__init__(self.name)


    def _validate(self):
        if self.relative_size_int < -4 or self.relative_size_int > 5:
            raise ValueError('relative size int must be in the range of -4 to 5')


def latex_size_str_for_relative_size_int(relative_size_int: int) -> str:
    latex_size_index = relative_size_int + 4
    return LATEX_SIZES[latex_size_index]



from pyexlatex.models.section.base import SectionBase
from pyexlatex.models.section.paragraphs import Paragraph


class SubSubSection(SectionBase):
    """
    A part of a subsection, the third largest section type.
    """
    name = 'subsubsection'
    next_level_down_class = Paragraph


class SubSection(SectionBase):
    """
    A part of a section, the second largest section type.
    """
    name = 'subsection'
    next_level_down_class = SubSubSection


class Section(SectionBase):
    """
    A section of the document, the largest section type.
    """
    name = 'section'
    next_level_down_class = SubSection








from copy import deepcopy

from pyexlatex.texgen import _toprule_str, _midrule_str, _bottomrule_str, _cmidrule_str
from pyexlatex.models.mixins import StringAdditionMixin
from mixins.repr import ReprMixin
from pyexlatex.table.models.mixins.addvalues.row import RowAddMixin


class TableLine(StringAdditionMixin, ReprMixin):
        pass

class TopRule(TableLine):

    def __str__(self):
        return _toprule_str()


class MidRule(TableLine):

    def __str__(self):
        return _midrule_str()

class BottomRule(TableLine):

    def __str__(self):
        return _bottomrule_str()

class TableLineSegment(StringAdditionMixin, ReprMixin, RowAddMixin):
    repr_cols = ['col_from', 'col_to']

    def __init__(self, col_from: int, col_to: int=None, align: str='lr'):
        self.col_from = col_from
        self.col_to = col_to
        self.align = align

    def __str__(self):
        return _cmidrule_str(self.align, self._range_str)

    def __len__(self):
        if self.col_to is None:
            return 1

        return (self.col_to - self.col_from) + 1

    def shift(self, shift: int):
        """
        Increments col_from and col_to by int. Used to shift underlines as tables are combined
        :return:
        """
        self.col_from += shift
        self.col_to += shift

    @property
    def _range_str(self):
        # add 1 to all columns to reindex to latex indices (1-based) from python indices (0-based)

        if self.col_to is None:
            col_to = self.col_from # single column
        else:
            col_to = self.col_to # multi column

        return f'{self.col_from + 1}-{col_to + 1}'


    def _add_class(self, other):
        return TableLineOfSegments

    @classmethod
    def from_list_of_ints(cls, int_list: [int]):
        if len(int_list) == 1:
            return cls(int_list[0])
        else:
            return cls(int_list[0], int_list[-1])

    @property
    def is_spacer(self):
        return True

class TableLineOfSegments(RowAddMixin, TableLine):
    repr_cols = ['values', 'num_columns']

    def __init__(self, segments: [TableLineSegment], num_columns: int=None):
        self.values = segments
        self.num_columns = num_columns

    def __str__(self):
        return ' '.join(str(segment) for segment in self.values)

    def __len__(self):
        return self.num_columns

    def __add__(self, other):
        # need to handle shifting of indices
        if isinstance(other, (TableLineOfSegments, TableLineSegment)):
            to_add = deepcopy(other)
            to_add.shift(len(self))
        else:
            raise NotImplementedError(f'cannot add type {type(other)} to TableLineOfSegments')
        line_of_segments: TableLineOfSegments =  super().__add__(to_add)
        line_of_segments.num_columns = self.num_columns + to_add.num_columns
        return line_of_segments

    def __radd__(self, other):
        # need to handle shifting of indices
        try:
            shift = len(other)
        except TypeError:
            raise NotImplementedError('cannot right add TableLineOfSegments without knowing how many columns to shift. '
                                      f'Could not calculate length of other. type of other: {type(other)}')
        to_add = deepcopy(self)
        to_add.shift(shift)
        line_of_segments: TableLineOfSegments = RowAddMixin.radd(to_add, other)
        line_of_segments.num_columns = self.num_columns + other.num_columns
        return line_of_segments

    def __getitem__(self, item):
        return self.values[item]

    def __iter__(self):
        for value in self.values:
            yield value

    @property
    def num_columns(self):
        return self._num_columns

    @num_columns.setter
    def num_columns(self, num_columns):
        if num_columns is None:
            self._num_columns = 0
        elif isinstance(num_columns, int):
            self._num_columns = num_columns
        else:
            raise ValueError(f'must provide int or None for num columns in TableLineOfSegments. got {num_columns}')

    def shift(self, shift):
        """
        Increments col_from and col_to by int for all segments within this line of segments
        :param shift:
        :return:
        """
        #### TEMP
        # import pdb
        # pdb.set_trace()
        #### END TEMP
        [value.shift(shift) for value in self.values]

    def _add_class(self, other):
        return TableLineOfSegments

    @classmethod
    def from_list_of_ints(cls, int_list: [int], num_columns: int=None):
        if len(int_list) == 1:
            return cls([TableLineSegment(int_list[0])])

        # create ranges out of consecutive ints. Add segments with either ranges or individual ints
        segments = []
        current_range = []
        for i, int_ in enumerate(int_list):
            if i != 0:
                # now have int and last int for all loops
                if int_ - last_int == 1:  # consecutive range continues
                    current_range.append(int_)
                else:  # break in range. need to output prior, and start new range
                    segments.append(TableLineSegment.from_list_of_ints(current_range))
                    current_range = [int_]

            last_int = int_
            if i == 0: # initialization
                current_range.append(int_)

        segments.append(TableLineSegment.from_list_of_ints(current_range))

        return cls(segments, num_columns=num_columns)

    def pad(self, length: int, direction='right'):
        """
        Method is here to avoid errors when adding rows together through TableSection.__add__
        No padding is actually necessary for these rows as they are not in the tabular format,
        these lines only specify the columns which need to be underlined.

        However when padding left, we just need to shift the values rather than creating padding
        """
        direction = direction.lower().strip()
        if direction == 'right':
            pass # no shift necessary, column numbers have not changed
        elif direction == 'left':
            self.shift(length - len(self)) # shift column indices right to adjust for new left cells
        else:
            raise ValueError(f'must pass right or left for direction. got {direction}')

        self.num_columns = length # extend number of columns to new length

    @property
    def is_spacer(self):
        return all([value.is_spacer for value in self.values])


import os


def latex_filename_replacements(string):
    return string.replace('%', 'pct').replace('/','_').replace('\\','_').replace('-','_').replace('$','')


def _latex_valid_basename(filepath):
    basename = os.path.basename(filepath)
    return basename.replace(' ', '_').replace('/','_').replace('%','pct').replace('$','')


import os
from tempdir import TempDir

from pyexlatex.tools import date_time_move_latex
from pyexlatex.logic.pdf.fileops import _move_if_exists_and_is_needed, _copy_if_needed
from pyexlatex.texgen.replacements.filename import _latex_valid_basename
from pyexlatex.logic.pdf.api.main import latex_str_to_pdf_obj
from pyexlatex.typing import BytesListOrNone, StrList, BytesList, StrListOrNone


def document_to_pdf_and_move(document, outfolder, image_paths: StrListOrNone = None, outname='figure', as_document=True,
                             move_folder_name='Figures', image_binaries: BytesListOrNone = None,
                             run_bibtex: bool = False, date_time_move: bool = False):

    # Create tex file
    outname_tex = outname + '.tex'
    outpath_tex = os.path.abspath(os.path.join(outfolder, outname_tex))
    with open(outpath_tex, 'w', encoding='utf8') as f:
        f.write(str(document))

    tex_inputs = output_sources_return_tex_input_paths(
        outfolder, image_paths=image_paths, image_binaries=image_binaries
    )

    if as_document:
        outname_pdf = outname + '.pdf'
        outpath_pdf = os.path.abspath(os.path.join(outfolder, outname_pdf))
        latex_str_to_pdf_file(str(document), outpath_pdf, texinputs=tex_inputs, run_bibtex=run_bibtex)
    if not date_time_move:
        return

    # Handle date/time move functionality
    new_outfolder = date_time_move_latex(outname, outfolder, folder_name=move_folder_name) #move table into appropriate date/number folder

    if image_paths and new_outfolder:
        # Copy second time to move pictures along with pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        sources_outfolder = os.path.join(new_outfolder, 'Sources')
        if not os.path.exists(sources_outfolder):
            os.makedirs(sources_outfolder)
        if image_binaries:
            _write_image_paths_and_binaries_to_folder(sources_outfolder, image_paths, image_binaries)
        else:
            [_move_if_exists_and_is_needed(
                os.path.join(sources_tempfolder, _latex_valid_basename(filepath)),
                os.path.join(sources_outfolder, _latex_valid_basename(filepath))
             )
            for filepath in image_paths]


def latex_file_to_pdf(folder: str, filename: str):
    # create PDF. Need to run twice for last page, as is written to aux file on the first iteration and
    # aux file is used on the second iteration
    orig_path = os.getcwd()
    os.chdir(folder)
    pdflatex_command = f'pdflatex "{filename}"'
    [os.system(pdflatex_command) for _ in range(2)]
    os.chdir(orig_path)


def latex_str_to_pdf_file(latex_str: str, filepath: str, texinputs: StrListOrNone = None,
                          run_bibtex: bool = False):
    pdf = latex_str_to_pdf_obj(latex_str, texinputs=texinputs, run_bibtex=run_bibtex)
    pdf.save_to(filepath)
    return pdf


def latex_str_to_pdf_obj_with_sources(latex_str: str, image_paths: StrListOrNone = None,
                                      image_binaries: BytesListOrNone = None, run_bibtex: bool = False):
    with TempDir() as tmpdir:
        tex_inputs = output_sources_return_tex_input_paths(
            tmpdir, image_paths=image_paths, image_binaries=image_binaries
        )
        pdf = latex_str_to_pdf_obj(latex_str, texinputs=tex_inputs, run_bibtex=run_bibtex)

    return pdf


def output_sources_return_tex_input_paths(outfolder: str, image_paths: StrListOrNone = None,
                                          image_binaries: BytesListOrNone = None):
    if image_paths:
        # Copy first time for creation of pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        tex_inputs = [os.path.abspath(outfolder), os.path.abspath(sources_tempfolder), '.']
        if not os.path.exists(sources_tempfolder):
            os.makedirs(sources_tempfolder)
        if image_binaries:
            _write_image_paths_and_binaries_to_folder(sources_tempfolder, image_paths, image_binaries)
        else:
            [_copy_if_needed(filepath, os.path.join(sources_tempfolder, _latex_valid_basename(filepath)))
             for filepath in image_paths]
    else:
        tex_inputs = None

    return tex_inputs

def _write_image_paths_and_binaries_to_folder(folder: str, image_paths: StrList, image_binaries: BytesList):
    if len(image_binaries) != len(image_paths):
        raise ValueError('must have equal image_binaries and image_path lengths if image_binaries are passed')
    for filepath, binary in zip(image_paths, image_binaries):
        image_outpath = os.path.join(folder, _latex_valid_basename(filepath))
        with open(image_outpath, 'wb') as f:
            f.write(binary)





def format_contents(contents) -> str:
    from pyexlatex.logic.builder import _build
    if isinstance(contents, (list, tuple)):
        return _build([_format_content(c) for c in contents])

    return _format_content(contents)


def _format_content(content) -> str:
    from pyexlatex.texgen.replacements.file import general_latex_replacements
    if isinstance(content, str):
        return general_latex_replacements(str(content))
    else:
        # Class is responsible for formatting. This may be a latex class or some
        # other harmless conversion such as int. It may also be an issue if the __str__
        # method of the class is not valid latex
        return content



from typing import Optional, Dict
from pyexlatex.models.references.bibtex.base import BibTexEntryBase


class BibTexArticle(BibTexEntryBase):
    """
    Biblography document which is an article.
    """
    item_type = 'article'
    required_attrs = ['author', 'title', 'journal', 'year']
    optional_attrs = ['volume', 'number', 'pages', 'month', 'note']

    def __init__(self, item_accessor: str, author: str, title: str, journal: str, year: str, volume: Optional[str] = None,
                 number: Optional[str] = None, pages: Optional[str] = None, month: Optional[str] = None,
                 note: Optional[str] = None):
        self.author = author
        self.title = title
        self.journal = journal
        self.year = year
        self.volume = volume
        self.number = number
        self.pages = pages
        self.month = month
        self.note = note

        super().__init__(item_accessor)



from pyexlatex.models.environmenttemplate import EnvironmentTemplate


class NoPageBreak(EnvironmentTemplate):
    """
    Contents within this environment will not have a page break come in the middle. Contents will instead be moved
    onto the next page if a break would have occurred.
    """
    name = 'absolutelynopagebreak'

    # taken from https://tex.stackexchange.com/a/94702
    begin_def = r'\par\nobreak\vfil\penalty0\vfilneg\vtop\bgroup'
    end_def = r'\par\xdef\tpd{\the\prevdepth}\egroup\prevdepth=\tpd'


from typing import Iterable, Any

# TODO: refactor to simplify

class CouldNotExtractObjsWithAttrException(Exception):
    pass

def extract_objs_with_attr_from_ambiguous_collection(collection, attr: str, attr_value: Any):
    if hasattr(collection, attr) and getattr(collection, attr) == attr_value:
        return collection
    if isinstance(collection, dict):
        return _extract_objs_with_attr_from_dict(collection, attr, attr_value)
    elif isinstance(collection, Iterable):
        return _extract_objs_of_type_from_normal_iterable(collection, attr, attr_value)
    else:
        raise CouldNotExtractObjsWithAttrException(f'could not extract objs from {collection} with attr {attr}={attr_value}')


def _extract_objs_with_attr_from_dict(collection, attr: str, attr_value: Any):
    collected_objs = []
    for key, obj in collection.items():
        if hasattr(obj, attr) and getattr(obj, attr) == attr_value:
            collected_objs.append(obj)
        elif isinstance(obj, dict):
            collected_objs += _extract_objs_with_attr_from_dict(collection, attr, attr_value)
        elif isinstance(obj, Iterable):
            collected_objs += _extract_objs_of_type_from_normal_iterable(obj, attr, attr_value)
        # else, skip the object

    return collected_objs


def _extract_objs_of_type_from_normal_iterable(collection, attr: str, attr_value: Any):
    collected_objs = []
    for obj in collection:
        if hasattr(obj, attr) and getattr(obj, attr) == attr_value:
            collected_objs.append(obj)
        elif isinstance(obj, dict):
            collected_objs += _extract_objs_with_attr_from_dict(collection, attr, attr_value)
        elif isinstance(obj, Iterable):
            collected_objs += _extract_objs_of_type_from_normal_iterable(obj, attr, attr_value)
        # else, skip the object

    return collected_objs



from pyexlatex.models.sizes.textwidth import TextWidth
from pyexlatex.models.format.rule import Rule


class HLine(Rule):
    """
    Draws a horizontal line across the text width.
    """

    def __init__(self, thickness: float = 0.4):
        super().__init__(length=TextWidth(), thickness=thickness)



from pyexlatex.models.item import NoOptionsNoContentsItem


class And(NoOptionsNoContentsItem):
    """
    Used internally for writing multiple institutions.
    """
    name = 'and'

    def __init__(self):
        super().__init__(self.name)


import os
from typing import Union, List, Dict, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from matplotlib.pyplot import Axes, Figure as PltFigure

from pyexlatex.figure.models.subfigure import Subfigure, Graphic
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models import Item
from pyexlatex.models.caption import Caption
from pyexlatex.models.label import Label
from pyexlatex.models.landscape import Landscape
from pyexlatex.logic.builder import build_figure_content
from pyexlatex.texgen.replacements.filename import latex_filename_replacements

from pyexlatex.models.commands.newenvironment import NewEnvironment
from pyexlatex.models.commands.begin import Begin
from pyexlatex.models.commands.end import End
from pyexlatex.models.environment import Environment
from pyexlatex.models.containeritem import ContainerItem

SubfigureOrGraphic = Union[Subfigure, Graphic]
SubfiguresOrGraphics = List[SubfigureOrGraphic]
PltFigureOrAxes = Union['Axes', 'PltFigure']
PltFigureOrAxesNameDict = Dict[str, PltFigureOrAxes]

class Figure(ContainerItem, Item):
    """
    used for creating latex figures from images. Currently the main usage is the Figure class created with the method
    Figure.from_dict_of_names_and_filepaths. Pass a dictionary where the keys are names for subfigures and the values
    are filepaths where the image for the subfigure is located.

    """
    name = 'figure'

    def __init__(self, subfigures: SubfiguresOrGraphics, caption=None, label=None, centering=True, position_str=None,
                 landscape: bool=False):
        self.subfigures = subfigures
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None
        self.centering = centering
        self.landscape = landscape

        self._remove_subfigure_elevate_contents_to_figure_if_single_subfigure()

        self.add_data_from_content(self.subfigures)

        content = build_figure_content(
            self.subfigures,
            caption=self.caption,
            label=self.label,
            centering=self.centering,
            position_str=position_str
        )

        super().__init__(self.name, content)

        if landscape:
            lfigure_def = NewEnvironment(
                'lfigure',
                Begin('landscape') + Begin('figure'),
                End('figure') + End('landscape')
            )
            self.data.begin_document_items.append(lfigure_def)
            self.env = Environment('lfigure')

    def __repr__(self):
        return f'<Figure(subfigures={self.subfigures}, caption={self.caption})>'

    def __iter__(self):
        for subfigure in self.subfigures:
            yield subfigure

    def __getitem__(self, item):
        return self.subfigures[item]

    def as_document(self, landscape=False):
        from pyexlatex.models.document import Document
        from pyexlatex.figure.packages import default_packages

        return Document(self, default_packages, landscape=landscape)

    def to_pdf_and_move(self, as_document=True, outfolder: str=None, outname: str=None,
                        landscape=False):
        from pyexlatex.logic.pdf.main import document_to_pdf_and_move
        from pyexlatex.models.document import Document

        to_output: Figure = self
        if as_document:
            to_output: Document = self.as_document(
                landscape=landscape if self.landscape == False else False  # don't apply landscape twice
            )

        if outfolder is None:
            outfolder = '.'
        if outname is None:
            outname = 'figure'
        else:
            outname = latex_filename_replacements(outname)

        document_to_pdf_and_move(
            to_output,
            outfolder,
            image_paths=self.data.filepaths,
            outname=outname,
            as_document=as_document,
            image_binaries=self.data.binaries
        )


    @classmethod
    def from_dict_of_names_and_filepaths(cls, filepath_name_dict: dict, figure_name: str=None,
                                         position_str_name_dict: dict=None, label=None, centering=True,
                                         landscape: bool = False):
        """

        Args:
            filepath_name_dict: dictionary where keys are names of subfigures and values
                                are the filepaths to the images for those subfigures.
            figure_name: name for overall figure
            position_str_name_dict: dictionary where keys are names of subfigures and values
                                are the position strs for those figures, e.g. r'[t]{0.45\linewidth}'

        Returns:

        """

        # TODO: add possibility of passing grid shape rather than actual position str

        if position_str_name_dict is None:
            position_str_name_dict = {}

        subfigures = []
        for name, filepath in filepath_name_dict.items():
            subfigures.append(
                Subfigure(
                    filepath,
                    caption=name,
                    position_str=position_str_name_dict[name] if name in position_str_name_dict else r'[t]{0.45\linewidth}'
                )
            )

        return cls(
            subfigures,
            caption=figure_name,
            label=label,
            centering=centering,
            landscape=landscape
        )

    @classmethod
    def from_dict_of_names_and_plt_figures(cls, plt_fig_name_dict: PltFigureOrAxesNameDict, sources_outfolder: str,
                                           source_filetype: str = 'pdf',
                                           figure_name: str=None,
                                           position_str_name_dict: dict=None, label=None, centering=True,
                                           landscape: bool = False):
        """

        Args:
            plt_fig_name_dict: Key is display name in output figure, value is matplotlib axes or figure
            sources_outfolder: folder to output individual plt figures
            source_filetype: Filetype for individual plt figures. The default is pdf. Use png or another image type if
                outputting complicated figures or performance may be affected when viewing the pdf.
            figure_name: name for overall figure
            position_str_name_dict: dictionary where keys are names of subfigures and values
                are the position strs for those figures, e.g. r'[t]{0.45\linewidth}'
            label:
            centering:
            landscape:

        Returns: Figure

        """

        filepath_name_dict = {}  # store outputted filepaths of sources to pass to from_dict_of_names_and_filepaths
        for name, plt_figure_or_axes in plt_fig_name_dict.items():
            plt_figure = _get_plt_figure_from_axes_or_figure(plt_figure_or_axes)
            outpath = os.path.join(sources_outfolder, f'{latex_filename_replacements(name)}.{source_filetype}')
            plt_figure.savefig(outpath)
            filepath_name_dict[name] = outpath

        return cls.from_dict_of_names_and_filepaths(
            filepath_name_dict,
            figure_name=figure_name,
            position_str_name_dict=position_str_name_dict,
            label=label,
            centering=centering,
            landscape=landscape
        )



    def _remove_subfigure_elevate_contents_to_figure_if_single_subfigure(self):
        """
        If there is a single subfigure, leaving in subfigure format results in very small output. Must strip
        out subfigure layer, leaving only a figure with a graphic, then it can fill the page.
        :return:
        """
        if len(self.subfigures) != 1:
            return

        self.subfigures: List[Union[Subfigure, Graphic]]

        if hasattr(self.subfigures[0], 'graphic'):
            # got subfigure
            item_is_subfigure = True
            orig_graphic = self.subfigures[0].graphic
        else:
            # got graphic
            item_is_subfigure = False
            orig_graphic = self.subfigures[0]

        # Elevate caption of sub-figure if there is no figure caption
        if self.caption is None and item_is_subfigure:
            self.caption = self.subfigures[0].caption

        # update width to whole page
        graphic = Graphic(orig_graphic.filepaths[0], width=r'1.1\paperwidth')

        # need to turn off centering to cover whole page
        self.centering = False

        self.subfigures = [graphic]


def _get_plt_figure_from_axes_or_figure(plt_axes_or_fig: PltFigureOrAxes) -> 'PltFigure':
    # Both axes and figure have the get_figure method, however for the figure, it will return None
    possible_figure = plt_axes_or_fig.get_figure()
    if possible_figure is None:
        return plt_axes_or_fig  # had figure already
    else:
        return possible_figure # extracted figure from axes






from typing import Union

from pyexlatex.table.models.table.row import Row
from pyexlatex.table.models.labels.collection import LabelCollection, Label


class LabelRow(Row):
    repr_cols = ['values', 'length']

    def __init__(self, values: Union[LabelCollection, Label]):
        super().__init__(values)

    @property
    def length(self):
        return len(self)

    def _add_class(self, other):
        if isinstance(other, (LabelCollection, Label)):
            return LabelRow
        else:
            return super()._add_class(other)



from typing import Optional, Sequence
from copy import deepcopy

from pyexlatex.table.models.table.row import Row
from pyexlatex.logic.tools import _max_len_or_zero
from mixins.repr import ReprMixin


class TableSection(ReprMixin):
    repr_cols = ['rows']

    def __init__(self, rows: Sequence[Row], break_size_adjustment: Optional[str] = None):
        self.break_size_adjustment = break_size_adjustment
        self.rows = rows

    def __iter__(self):
        for row in self.rows:
            yield row

    def __getitem__(self, item):
        return self.rows[item]

    @property
    def length(self):
        return len(self.rows)

    def __str__(self) -> str:
        from pyexlatex.table.logic.table.build import _build_tabular_str_from_rows_and_lines
        return _build_tabular_str_from_rows_and_lines(self.rows, self.break_size_adjustment)

    def __add__(self, other):
        # import here to avoid circular imports
        from pyexlatex.table.models.spacing.columntable import ColumnPadTable

        num_rows = max([self.length, other.length])

        out_rows = []

        for row_num in range(num_rows):
            out_row_class = _get_class_for_row(self, other, row_number=row_num)
            out_row = out_row_class([])
            try:
                out_row += self[row_num]
            except IndexError:
                # expected to hit here when sections have different numbers of rows
                out_row.pad(self.num_columns, direction='left')
            # Special handling for ColumnPadTable. Ignore rows, just always pad right
            if isinstance(other, ColumnPadTable):
                out_row.pad(self.num_columns + other.width, direction='right')
            else:
                try:
                    out_row += other[row_num]
                except IndexError:
                    out_row.pad(other.num_columns + self.num_columns, direction='right')

            out_rows.append(out_row)

        klass = self._add_class(other)

        return klass(out_rows)

    def __radd__(self, other):
        num_rows = max([self.length, other.length])

        out_rows = []

        for row_num in range(num_rows):
            out_row_class = _get_class_for_row(self, other, row_number=row_num)
            out_row = out_row_class([])
            try:
                out_row += other[row_num]
            except IndexError:
                out_row.pad(other.num_columns, direction='left')
            try:
                out_row += self[row_num]
            except IndexError:
                # expected to hit here when sections have different numbers of rows
                out_row.pad(self.num_columns + other.num_columns, direction='right')

            out_rows.append(out_row)

        klass = self._add_class(other)

        return klass(out_rows)

    def _add_class(self, other):
        # keep same class if both are same class
        # otherwise, default to Row class
        self_class = type(self)
        other_class = type(other)
        klass = self_class if self_class == other_class else TableSection

        # import here to avoid circular imports
        from pyexlatex.table.models.spacing.rowtable import RowPadTable
        from pyexlatex.table.models.spacing.columntable import ColumnPadTable
        if klass in (RowPadTable, ColumnPadTable):
            return TableSection

        return klass

    @property
    def num_columns(self):
        try:
            return self._num_columns
        except AttributeError:
            self._num_columns = self._set_num_columns()

        return self._num_columns

    def _set_num_columns(self):
        return _max_len_or_zero(self.rows)

    def join(self, sections):
        """
        Repliactes str.join behavior. Useful for creating padding spaces in a PanelGrid/PanelCollection
        :param sections:
        :return:
        """
        if len(sections) == 1:
            return sections[0]

        for i, section in enumerate(sections):
            if i == 0:
                out_section = deepcopy(section)
                continue
            else:
                out_section += self # insert self inbetween sections, replicating join behavior
            out_section = out_section + section

        return out_section

    def pad(self, length: int, direction='right'):
        """
        Expand table out to the right or left with blanks, until it is length passed (apply to every row)
        :param length:
        :param direction:
        :return:
        """
        [row.pad(length=length, direction=direction) for row in self.rows]

    @property
    def is_spacer(self):
        return all([row.is_spacer for row in self.rows])


    def _recreate_rows_if_created(self):
        """
        Used for recreating rows after labels or value table have been modified.
        Don't want to create rows if they've never been created before (initialization)
        :return:
        """
        # run creation process again
        if hasattr(self, '_rows'):
            self._rows = self._create_rows()
        # else, do nothing

    def _create_rows(self):
        raise NotImplementedError

def _get_class_for_row(*objs, row_number: int=0):
    from pyexlatex.table.models.texgen.lines import TableLineOfSegments
    from pyexlatex.table.models.labels.row import LabelRow, LabelCollection

    suggested_class = _get_by_row_number_first_class_without_index_error(*objs, row_number=row_number)

    # this is one row class that should not be created unless explicitly calling the rows method of LabelTable.
    # this is because LabelCollections contain underlines, which are created separately from LabelRows in the
    # _create_rows method of LabelTable. If adding a collection to a row, will lose the underline
    if issubclass(suggested_class, LabelRow):
        return LabelCollection

    # already got a row class, just return it
    if issubclass(suggested_class, (Row, TableLineOfSegments)):
        return suggested_class

    # otherwise, we have an item class. return the appropriate row class for that item
    return _get_row_class_for_item_class(suggested_class)


def _get_row_class_for_item_class(klass):
    from pyexlatex.table.models.texgen.lines import TableLineSegment, TableLineOfSegments
    from pyexlatex.table.models.data.dataitem import DataItem
    from pyexlatex.table.models.data.row import DataRow
    from pyexlatex.table.models.labels.label import Label
    from pyexlatex.table.models.labels.row import LabelRow, LabelCollection
    if issubclass(klass, TableLineSegment):
        return TableLineOfSegments
    if issubclass(klass, DataItem):
        return DataRow
    if issubclass(klass, (Label, LabelCollection)):
        return LabelCollection
    else:
        raise NotImplementedError(f'could not determine row class for item class {klass}')

def _get_by_row_number_first_class_without_index_error(*objs, row_number: int):
    attrs_or_none = _get_by_row_number_none_if_index_error(*objs, row_number=row_number)
    return _first_non_none_class(attrs_or_none)

def _get_by_row_number_none_if_index_error(*objs, row_number: int):
    out_attrs = []
    for obj in objs:
        try:
            out_attrs.append(obj[row_number])
        except IndexError:
            out_attrs.append(None)
    return out_attrs

def _first_non_none_class(objs):
    valid_objs = [obj for obj in objs if obj is not None]
    if len(valid_objs) == 0:
        raise ValueError('all objects passed were None. could not determine which class to use.')

    return type(valid_objs[0])



from typing import List, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.logic.pdf.errors.exc import LatexException
from pyexlatex.logic.pdf.errors.models.error import LatexError
from pyexlatex.logic.pdf.errors.models.error_collection import LatexErrorCollection
from latex.exc import LatexBuildError


class LatexExceptionManager:

    def __init__(self, exceptions: List[type]):
        self.exceptions_list = exceptions
        self.exceptions = self._create_map()
        self.match_strs = [key for key in self.exceptions.keys()]

    def _create_map(self) -> Dict[str, 'LatexException']:
        exception_map = {}
        for exc in self.exceptions_list:
            exception_map[exc._match_str] = exc
        return exception_map

    def exception_from_error(self, error: LatexError) -> 'LatexException':
        from pyexlatex.logic.pdf.errors.exc import LatexException
        for possible_match in self.match_strs:
            if possible_match in error.error:
                exc_class = self.exceptions[possible_match]
                return exc_class(error.error)

        # No match found, return general exception
        return LatexException(error.error)

    def exceptions_from_latex_build_error(self, build_error: LatexBuildError) -> List['LatexException']:
        errors = LatexErrorCollection(build_error)
        return [self.exception_from_error(latex_error) for latex_error in errors]



"""
Create LaTeX documents using only Python. Rather than building a direct Python API to LaTeX, this package
has its own, simpler API to creating documents. It is focused on creating professional-looking documents
with little styling effort.
"""
from pyexlatex.models.document import Document
from pyexlatex.models.standalone import Standalone
from pyexlatex.models.equation import Equation
from pyexlatex.models.section.paragraphs import Paragraph, SubParagraph
from pyexlatex.models.section.sections import Section, SubSection, SubSubSection
from pyexlatex.models.section.appendix import Appendix
from pyexlatex.models.lists.unordered import UnorderedList
from pyexlatex.models.lists.ordered import OrderedList
from pyexlatex.models.format.raw import Raw
from pyexlatex.models.format.text.text import Text
from pyexlatex.models.ref import Ref, NameRef
from pyexlatex.models.footnote import Footnote
from pyexlatex.models.references.bibtex.article import BibTexArticle
from pyexlatex.models.references.bibtex.misc import BibTexMisc
from pyexlatex.models.references.bibtex.manual import BibTexManual
from pyexlatex.models.references.bibliography import Bibliography
from pyexlatex.models.references.citations import Cite, CiteP, CiteT
from pyexlatex.models.format.text.color.main import TextColor
from pyexlatex.models.format.text.monospace import Monospace
from pyexlatex.models.format.text.underline import Underline
from pyexlatex.models.format.text.bold import Bold
from pyexlatex.models.format.text.italics import Italics
from pyexlatex.models.format.text.smallcaps import SmallCaps
from pyexlatex.models.format.code.python import Python
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.fills import VFill, HFill
from pyexlatex.models.hyperlinks import Hyperlink
from pyexlatex.models.package import Package
from pyexlatex.models.page.header import Header
from pyexlatex.models.page.number import ThisPageNumber
from pyexlatex.models.format.vspace import VSpace
from pyexlatex.models.format.hspace import HSpace
from pyexlatex.models.format.paragraphindent import ParagraphIndent
from pyexlatex.models.format.hline import HLine
from pyexlatex.models.format.text.color.deftypes.rgb import RGB
from pyexlatex.models.format.text.color.deftypes.hex import Hex
from pyexlatex.models.template import Template
from pyexlatex.models.environmenttemplate import EnvironmentTemplate
from pyexlatex.models.format.breaks import OutputLineBreak
from pyexlatex.models.latex import LaTeX
from pyexlatex.models.format.nopagebreak import NoPageBreak
from pyexlatex.models.format.nolinebreak import NoLineBreak
from pyexlatex.models.format.paragraph.justifying import Justifying
from pyexlatex.models.format.centering import Center
from pyexlatex.models.control.newpage import PageBreak
from pyexlatex.models.control.group import Group
from pyexlatex.models.sizes.textsizes import TextSize



from typing import List, Union, Sequence, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay import Overlay
import posixpath
import os

from pyexlatex.models.item import SimpleItem
from pyexlatex.models.sizes.textwidth import TextWidth


class Graphic(SimpleItem):
    """
    Basic class for including graphics.

    In most documents, a Figure would be used to give more structure to the graphic. This class will literally just
    include the image with no other structure. It is typically more useful in presentations where the structure
    is already being provided.
    """
    name = 'includegraphics'

    def __init__(self, filepath, width: Union[str, float] = 1.0, cache: bool = True, options: Sequence[str] = None,
                 overlay: Optional['Overlay'] = None):
        """

        Args:
            filepath:
            width: if a float is passed, is interpreted as a fraction of line width. if a str is passed, will be passed
                into latex directly
            cache:
        """
        self._set_path(filepath)
        self.width = width
        self.binaries = None

        self.options = self._get_list_copy_from_list_or_none(options)
        self.options.append(
            f'width={self.width_str}'
        )

        if cache:
            self._cache(filepath)

        super().__init__(
            self.name,
            self.source_paths[0],
            pre_modifiers=self.options_str,
            overlay=overlay,
            format_content=False
        )



    def __repr__(self):
        return f'<Graphic({self.filepath}, width={self.width})>'


    def _set_path(self, filepath: str):
        from pyexlatex.texgen.replacements.filename import _latex_valid_basename

        basename = _latex_valid_basename(filepath)
        source_path = posixpath.join('Sources', basename)

        self._filepath_parts = filepath.split(os.path.sep)
        self.source_paths = [source_path]

    def _cache(self, filepath: str):
        with open(filepath, 'rb') as f:
            self.binaries = [f.read()]

    @property
    def filepaths(self) -> List[str]:
        return [os.path.sep.join(self._filepath_parts)]

    @property
    def filepath(self) -> str:
        return self.filepaths[0]

    @property
    def source_path(self) -> str:
        return self.source_paths[0]

    @property
    def width_str(self) -> str:
        # Handle float being passed
        if isinstance(self.width, (float, int)):
            return f'{self.width}{TextWidth()}'

        # Handle manually passing a width string
        return self.width

    @property
    def options_str(self) -> Optional[str]:
        if self.options is None:
            return None

        return self._wrap_with_bracket(', '.join(self.options))



from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.models.section.sections import Section


class Appendix(TextAreaBase):
    """
    Content to be placed at the end, separate from the main document.
    """
    name = 'appendices'
    next_level_down_class = Section

    def __init__(self, contents):
        super().__init__(self.name, contents)



from pyexlatex.models.item import NoOptionsNoContentsItem


class VFill(NoOptionsNoContentsItem):
    """
    Vertical fill, put between items to put vertical spacing between them.
    """
    name = 'vfill'

    def __init__(self):
        super().__init__(self.name)


class HFill(NoOptionsNoContentsItem):
    """
    Horizontal fill, put between items to put horizontal spacing between them.
    """
    name = 'hfill'

    def __init__(self):
        super().__init__(self.name)


from typing import Optional, Sequence

from pyexlatex.models.item import SimpleItem
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.control.documentclass.classtypes.manager import DocumentClassTypesManager


class DocumentClass(ContainerItem, SimpleItem):
    name = 'documentclass'

    def __init__(self, document_type: str = 'article', font_size: Optional[float] = None,
                 num_columns: Optional[int] = None, options: Optional[Sequence[str]] = None):
        manager = DocumentClassTypesManager()
        self.document_type = manager.get_class_type_by_name(document_type)
        self.add_data_from_content(self.document_type)
        self.font_size = font_size
        self.num_columns = num_columns
        self.options = options
        super().__init__(self.name, document_type, pre_modifiers=self._pre_modifiers_str)

    @property
    def _pre_modifiers_str(self) -> str:
        options = self._get_list_copy_from_list_or_none(self.options)
        if self.font_size is not None:
            options.append(f'{self.font_size}pt')
        if self.num_columns is not None:
            options.append(self._num_columns_str)

        options_str = ', '.join(options)
        return self._wrap_with_bracket(options_str)

    def _validate(self):
        if self.font_size is not None:
            if self.font_size not in (10, 11, 12):
                raise NotImplementedError(f'need to add a package to use font size other than 10, 11, 12, '
                                          f'got {self.font_size}')
        if self.num_columns is not None:
            if self.num_columns not in (1, 2):
                raise NotImplementedError(f'need to use the multicol package to get other than 1, 2 columns, '
                                          f'got {self.num_columns}')

    @property
    def _num_columns_str(self) -> str:
        if self.num_columns == 1:
            return 'onecolumn'
        if self.num_columns == 2:
            return 'twocolumn'
        return ''


from typing import List, Optional, Sequence
from pyexlatex.models.document import DocumentBase
from pyexlatex.typing import ItemOrListOfItems
from pyexlatex.models.package import Package
from pyexlatex.models.control.documentclass.documentclass import DocumentClass


class Standalone(DocumentBase):
    """
    The main high-level class for creating one-off graphics, equations, etc. Does not provide any document structure
    and will crop the document to the contents.
    """
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package] = None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None,
                 font_size: Optional[float] = None, doc_class_options: Optional[Sequence[str]] = None):

        self.document_class_obj = DocumentClass(
            document_type='standalone',
            font_size=font_size,
            options=doc_class_options
        )

        super().__init__(
            content,
            packages=packages,
            pre_env_contents=pre_env_contents,
        )





from pyexlatex.models.item import SimpleItem
from pyexlatex.texgen import _todays_date_str

class Date(SimpleItem):
    name = 'date'

    def __init__(self, contents=None):
        if contents is None:
            contents = _todays_date_str()
        super().__init__(self.name, contents)


from typing import Sequence, Union, Optional, List
from copy import deepcopy
from pyexlatex.models.item import ItemBase
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.graphics.tikz.node.node import Node
from pyexlatex.graphics import Arrow
from pyexlatex.graphics.tikz.node.position.directions import Right, Below


class LinearFlowchart(ContainerItem, ItemBase):
    """
    Creates a linear flow chart (nodes with arrows in between), going horizontally or vertically from one item
    to the next.
    """

    def __init__(self, steps: Sequence[Union[Node, str]], horizontal: bool = True,
                 node_options: Optional[Sequence[str]] = None):
        self.steps = steps
        self.horizontal = horizontal
        self.node_options = node_options
        self.add_data_from_content(steps)
        self.nodes = self._get_nodes()
        self.contents = self._get_contents()

    def __str__(self) -> str:
        from pyexlatex.logic.builder import _build
        if isinstance(self.contents, (list, tuple)):
            return _build(self.contents)
        else:
            return str(self.contents)

    def _get_contents(self) -> List[Union[Node, Arrow]]:
        nodes = self.nodes
        contents = deepcopy(nodes)
        for i, node in enumerate(nodes):
            if i == 0:
                continue
            contents.append(Arrow(nodes[i - 1], nodes[i]))
        self.add_data_from_content(contents)
        return contents

    def _get_nodes(self) -> List[Node]:
        out_nodes = []
        for i, item in enumerate(self.steps):
            if hasattr(item, 'is_Node') and item.is_Node:
                item: Node
                if i == 0:
                    out_nodes.append(item)
                    continue
                # If beyond the first element, need to create a new node with the same info, but with
                # position relative to the last element
                new_node = Node(
                    contents=item.content,
                    location=self.direction(of=out_nodes[i - 1]),
                    label=item.label,
                    options=item.options,
                    overlay=item.overlay
                )
                new_node.add_data_from_content(item)
                out_nodes.append(
                    new_node
                )
            else:
                # Treat as str passed, need to create node with str and node options
                out_nodes.append(
                    Node(
                        contents=item,
                        location=self.direction(of=out_nodes[i - 1]) if i > 0 else None,
                        options=self.node_options
                    )
                )
        return out_nodes

    @property
    def direction(self) -> type:
        if self.horizontal:
            return Right
        else:
            return Below



from typing import List

from pyexlatex.logic.extract.by_type import extract_objs_of_type_from_ambiguous_collection
from pyexlatex.logic.extract.by_attr import extract_objs_with_attr_from_ambiguous_collection
from pyexlatex.models.documentitem import DocumentItem


def extract_document_items_from_ambiguous_collection(collection) -> List[DocumentItem]:
    return extract_objs_with_attr_from_ambiguous_collection(collection, 'is_DocumentItem', True)


def extract_document_items_from_ambiguous_collection_by_type(collection) -> List[DocumentItem]:
    return extract_objs_of_type_from_ambiguous_collection(collection, DocumentItem)


def is_latex_item(item) -> bool:
    return hasattr(item, 'is_LatexItem') and getattr(item, 'is_LatexItem')


from pyexlatex.models.item import SimpleItem


class PageStyle(SimpleItem):
    """
    Sets style of the pages for the entire document.
    """
    name = 'pagestyle'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class ThisPageStyle(SimpleItem):
    """
    Sets style of the current page in a document.
    """
    name = 'thispagestyle'

    def __init__(self, contents):
        super().__init__(self.name, contents)


from typing import Tuple, Union, Optional
from pyexlatex.graphics.tikz.node.position.directions import DirectionBase


class NodePosition:

    def __init__(self, location: Optional[Union[Tuple[int, int], DirectionBase, str]] = None):
        self.location = location
        self.absolute_location, self.relative_location = False, False
        if location is not None:
            if isinstance(location, (tuple, list)):
                self.absolute_location = True
            elif isinstance(location, str):
                # passing custom type of location, just wrap in parentheses
                self.location = f'({self.location})'
                self.absolute_location = True
            else:
                self.relative_location = True

    def __str__(self) -> str:
        if self.location is None:
            return ''

        if self.absolute_location:
            return f'at {self.location}'

        if self.relative_location:
            return str(self.location)


from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin


class Name(TextAreaMixin, SimpleItem):
    name = 'name'

    def __init__(self, author: str):
        self.author = author
        super().__init__(self.name, self.author)



from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.item import SimpleItem


class Bold(TextAreaMixin, SimpleItem):
    """
    Makes text bold.
    """
    name = 'textbf'

    def __init__(self, contents, **kwargs):
        super().__init__(self.name, contents, **kwargs)



from typing import Union, AnyStr


def _set_top_left_corner_labels(top_left_corner_labels = None):
    from pyexlatex.table import LabelTable, LabelCollection

    if top_left_corner_labels is None:
        return LabelTable.from_list_of_lists([[' ']])

    if isinstance(top_left_corner_labels, LabelTable):
        return top_left_corner_labels
    elif isinstance(top_left_corner_labels, LabelCollection):
        return LabelTable([top_left_corner_labels])
    elif isinstance(top_left_corner_labels, list):
        return LabelTable.from_list_of_lists([top_left_corner_labels]).T
    elif isinstance(top_left_corner_labels, str):
        return LabelTable.from_list_of_lists([[top_left_corner_labels]])
    else:
        raise NotImplementedError(f'was not able to create LabelTable out of {top_left_corner_labels}')


# This is the main settings file for package setup and PyPi deployment.
# Sphinx configuration is in the docsrc folder

# Main package name
PACKAGE_NAME = 'pyexlatex'

# Package version in the format (major, minor, release)
PACKAGE_VERSION_TUPLE = (0, 8, 10)

# Short description of the package
PACKAGE_SHORT_DESCRIPTION = 'Python Extends LaTeX - A High-Level Python API for Creating Latex Documents'

# Long description of the package
PACKAGE_DESCRIPTION = """
This project is aimed at creating LaTeX documents using only Python, without directly writing LaTeX code. Rather
than building a direct Python API to LaTeX, this package has its own, simpler API to creating documents. 
"""

# Author
PACKAGE_AUTHOR = "Nick DeRobertis"

# Author email
PACKAGE_AUTHOR_EMAIL = 'whoopnip@gmail.com'

# Name of license for package
PACKAGE_LICENSE = 'MIT'

# Classifications for the package, see common settings below
PACKAGE_CLASSIFIERS = [
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 3 - Alpha',

    # Indicate who your project is intended for
    'Intended Audience :: Developers',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'
]

# Add any third party packages you use in requirements here
PACKAGE_INSTALL_REQUIRES = [
    # Include the names of the packages and any required versions in as strings
    # e.g.
    # 'package',
    # 'otherpackage>=1,<2'
    'mixins',
    'pandas',
    'tempdir',
    'data',
    'latex',
    'future',
    'sympy',
    'shutilwhich',
    'bibtexparser'
]

# Sphinx executes all the import statements as it generates the documentation. To avoid having to install all
# the necessary packages, third-party packages can be passed to mock imports to just skip the import.
# By default, everything in PACKAGE_INSTALL_REQUIRES will be passed as mock imports, along with anything here.
# This variable is useful if a package includes multiple packages which need to be ignored.
DOCS_OTHER_MOCK_IMPORTS = [
    # Include the names of the packages as they would be imported, e.g.
    # 'package',
]

PACKAGE_URLS = {
    'Code': 'https://github.com/whoopnip/py-ex-latex/',
    'Documentation': 'https://whoopnip.github.io/py-ex-latex/'
}



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay import Overlay
from pyexlatex.models.item import NoBracesItem
from pyexlatex.models.section.base import TextAreaMixin
from mixins.repr import ReprMixin


class ListItem(TextAreaMixin, NoBracesItem, ReprMixin):
    name = 'item'
    repr_cols = ['contents']

    def __init__(self, contents, overlay: Optional['Overlay'] = None):
        super().__init__(self.name, contents, overlay=overlay)




from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models import Item
from pyexlatex.models.caption import Caption
from pyexlatex.models.label import Label
from pyexlatex.logic.builder import build_figure_content
from pyexlatex.models.containeritem import ContainerItem


class Subfigure(ContainerItem, Item):
    """
    Used for more control over building figures
    """
    name = 'subfigure'

    def __init__(self, filepath, caption=None, label=None, centering=True, position_str=r'[t]{0.45\linewidth}',
                 cache: bool = True):
        self.graphic = Graphic(filepath, cache=cache)
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None

        self.add_data_from_content(self.graphic)

        contents = build_figure_content(
            [self.graphic],
            caption=self.caption,
            label=self.label,
            centering=centering,
            position_str=position_str
        )

        super().__init__(self.name, contents)

    def __repr__(self):
        return f'<Subfigure({self.graphic.filepath}, caption={self.caption})>'






from typing import Optional
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.item import SimpleItem


# TODO: update all of this after getting a better understanding of how boxes work and the syntax.
# TODO: see: https://en.wikibooks.org/wiki/LaTeX/Boxes
class BoxBase(TextAreaMixin, SimpleItem):
    name = '<unimplemented, do not use BoxBase directly, use a subclass>'

    def __init__(self, contents, size_str: Optional[str]):
        super().__init__(self.name, contents, pre_modifiers=size_str, format_content=False)



from pyexlatex.models.item import SimpleItem


class NoHyphens(SimpleItem):
    """
    Prevents a line break within text, where a hyphen would normally be placed. However when it is used alone it
    will make text run over the margin. Use NoLineBreak to avoid this.
    """
    name = 'nohyphens'

    def __init__(self, contents, **kwargs):
        self.init_data()
        self.data.packages.append('hyphenat')
        super().__init__(self.name, contents, **kwargs)
